import os

class HandledError(Exception):
   pass

def tag_print(string: str, color: str = None, prompt: bool = False, end: str = "\n"):
	TAG = "[spotify-dlp] "
	string = TAG + string

	if color:
		string = color + string + Colors.ENDC

	return input(string) if prompt else print(string, end=end)

class Colors():
	ENDC = "\033[0m"
	FAIL = "\033[91m"
	WARN = "\033[93m"
	BOLD = "\033[1m"

class TokenFile():
	@staticmethod
	def get_token_filepath() -> str:
		if os.name == "nt":
			return os.path.join(os.getenv("APPDATA"), "spotify-dlp")
		else:
			return os.path.expanduser("~/.config/spotify-dlp/")

	@staticmethod
	def read_token(name: str) -> str:
		filepath = os.path.join(TokenFile.get_token_filepath(), name)

		if not os.path.exists(filepath):
			return None

		with open(filepath, "r") as f:
			return f.read().strip()

	@staticmethod
	def write_token(name: str, token: str):
		if not os.path.exists(TokenFile.get_token_filepath()):
			os.mkdir(TokenFile.get_token_filepath())

		with open(os.path.join(TokenFile.get_token_filepath(), name), "w+") as f:
			f.write(token)



from setuptools import setup, find_packages

setup(
	name="spotify-dlp",
	author="zWolfrost",
	version="2.4.0",
	description="Command line downloader for spotify tracks, playlists, albums and top artists songs.",
	long_description=open("README.md").read(),
	long_description_content_type="text/markdown",
	url="https://github.com/zWolfrost/spotify-dlp",
	packages=find_packages(),
	install_requires=[
		"requests>=2.31.0",
		"yt_dlp>=2024.4.9"
	],
	entry_points={
		"console_scripts": [
			"spotify-dlp = spotify_dlp.spotify_dlp:main"
		]
	}
)



import os, argparse, requests, yt_dlp
from spotify_dlp.spotify_api import SpotifyAPI, Item
from spotify_dlp.utils import HandledError, tag_print, Colors, TokenFile


def init_args() -> argparse.Namespace:
	parser = argparse.ArgumentParser(prog="spotify-dlp", description="Command line downloader for spotify tracks, playlists, albums and top artists tracks.")

	parser.add_argument("query", type=str, nargs=argparse.ZERO_OR_MORE, help="The words to search up or a link to a spotify album, artist, playlist or track. If \"saved\", download the user's saved tracks (requires browser authentication).")

	parser.add_argument("-a", "--auth", action="store_true", help="Authenticate using the client credentials flow and save the Client ID and Client Secret to a file.")
	parser.add_argument("-i", "--client-id", default=TokenFile.read_token("CLIENT_ID"), type=str, help="The Spotify Client ID.")
	parser.add_argument("-s", "--client-secret", default=TokenFile.read_token("CLIENT_SECRET"), type=str, help="The Spotify Client Secret.")

	parser.add_argument("-f", "--format", type=str, default="{title} - {authors} ({album})", help="The format of the downloaded tracks' names. Set to \"help\" for a list of available fields.")
	parser.add_argument("-t", "--type", type=str, default="track", choices=["album", "artist", "playlist", "track"], help="When searching up a query, the specified type of content.")
	parser.add_argument("-l", "--slice", type=str, default=":", help="The beginning and ending index of the list items to download separated by a colon \":\" (1-based). Either one of those indexes can be omitted.")

	parser.add_argument("-o", "--output", type=str, default=".", help="The output path of the downloaded tracks.")
	parser.add_argument("-c", "--codec", type=str, default="", choices=["m4a", "mp3", "flac", "wav", "aac", "ogg", "opus"], help="The audio codec of the downloaded tracks. By default, it is unchanged from the one \"yt-dlp\" downloads. Requires \"ffmpeg\" to be installed.")
	parser.add_argument("-m", "--metadata", action="store_true", help="Whether to download metadata (such as covers).")

	parser.add_argument("-y", "--yes", action="store_true", help="Whether to skip the confirmation prompt.")

	parser.add_argument("-v", "--verbose", action="store_true", help="Whether to display verbose information and full errors.")
	parser.add_argument("--version", action="version", version="%(prog)s 2.4.0")

	return parser.parse_args()

def parse_args(args: argparse.Namespace) -> argparse.Namespace:
	args.query = " ".join(args.query)

	try:
		begindex, endindex = args.slice.split(":")
	except ValueError:
		begindex, endindex = args.slice, args.slice

	try:
		begindex = 0    if (begindex == "" or begindex == "0") else int(begindex)-1
		endindex = None if (endindex == "" or endindex == "0") else int(endindex)

		args.slice = (begindex, endindex)
	except ValueError as e:
		raise HandledError("Invalid slice argument.") from e

	try:
		Item().format_with_index(args.format)
	except KeyError as e:
		raise HandledError(f"Invalid field \"{{{e.args[0]}}}\" in format argument. Use \"--format help\" to see available fields.") from e

	return args


def main():
	try:
		# Add color support for Windows terminals
		if os.name == "nt":
			os.system("")

		args = init_args()
		args = parse_args(args)


		### AUTHENTICATION ###
		if args.auth:
			print(
				"To authenticate, please follow these steps:\n"
				"1. Go to https://developer.spotify.com/dashboard/applications\n"
				"2. Press \"Create app\" and fill in details for name and description.\n"
				"3. Set \"Redirect URIs\" to a random URL, such as \"http://127.0.0.1:3000/\".\n"
				"4. Press \"Save\" and paste your Client ID and Client Secret below.\n"
			)
			client_id = input("Client ID: ").strip()
			client_secret = input("Client Secret: ").strip()

			if len(client_id) != 32 or not client_id.isalnum():
				raise HandledError("Invalid Client ID. Please try again.")

			if len(client_secret) != 32 or not client_secret.isalnum():
				raise HandledError("Invalid Client Secret. Please try again.")

			TokenFile.write_token("CLIENT_ID", client_id)
			TokenFile.write_token("CLIENT_SECRET", client_secret)
			tag_print(f"Tokens saved to ~/.config/spotify-dlp/", color=Colors.BOLD)

			return
		elif not args.query:
			raise HandledError("No query was provided. Please provide a query or a link to a Spotify album, artist, playlist or track.")

		if args.client_id and args.client_secret:
			try:
				spotify = SpotifyAPI.from_client_credentials_flow(args.client_id, args.client_secret)
			except Exception as e:
				raise HandledError("Couldn't fetch token. Client ID and/or Client Secret are probably invalid.") from e
		else:
			raise HandledError(
				"Not authenticated.\n"
				"You can authenticate by running `spotify-dlp --auth` and following the instructions.\n"
				"Alternatively, you can provide your Client ID and Client Secret as command line arguments."
			)


		### FETCH TRACKLIST ###

		try:
			try:
				SpotifyAPI.parse_url(args.query)
			except ValueError:
				tag_print("Searching up the query...")
				tracklist = spotify.items_by_search(args.query, args.type)
			else:
				tag_print("Fetching the query URL...")
				tracklist = spotify.items_by_url(args.query)
		except Exception as e:
			raise HandledError(e) from e

		if len(tracklist) == 0:
			raise HandledError("No tracks were found.")

		tracklist = tracklist[args.slice[0]:args.slice[1]]

		if len(tracklist) == 0:
			raise HandledError(f"The specified slice is out of range.")


		### DISPLAY TRACKLIST ###

		if args.format == "help":
			tag_print("Available fields for the format argument:", color=Colors.BOLD)
			for keys, value in tracklist[0].get_format_dict().items():
				print("{:>12} {}".format(f"{{{keys}}}:", value))
			return

		print()
		tag_print(f"The query you requested contained {len(tracklist)} track(s):", color=Colors.BOLD)
		for track in tracklist:
			print(track.format_with_index(args.format))

		if not args.yes:
			print()
			choice = tag_print("Are you sure you want to download these tracks? [Y/n]\n", color=Colors.BOLD, prompt=True)

			if "n" in choice.lower():
				return

			if choice:
				print()
		else:
			print()


		### DOWNLOAD TRACKS ###

		if not os.path.exists(args.output):
			os.makedirs(args.output)

		DEFAULT_YTDLP_OPTIONS = {
			"quiet": not args.verbose,
			"no_warnings": not args.verbose,
			"format": "bestaudio",
			"noplaylist": True,
			"extract_flat": True,
			"extractor_args": { "youtube": { "player_client": ["tv", "web"] } },
		} | ({
			"postprocessors": [
				{
					"key": "FFmpegExtractAudio",
					"preferredcodec": args.codec
				}
			]
		} if args.codec else {})

		for index, track in enumerate(tracklist, start=1):
			filename = track.safe_format(args.format)
			filepath = os.path.join(args.output, filename)

			if args.metadata:
				coverpath = os.path.join(args.output, f"{track.album}.jpg")
				if not os.path.exists(coverpath):
					try:
						img_data = requests.get(track.cover).content
						open(coverpath, "wb").write(img_data)
					except Exception as e:
						tag_print(f"An error was encountered while trying to download the cover for \"{track.album}\": {e}", color=Colors.WARN)
					else:
						if args.verbose:
							tag_print(f"Successfully downloaded the cover for \"{track.album}\"!")

			if os.path.exists(f"{filepath}.{args.codec}"):
				tag_print(f"File \"{filename}\" already exists; Skipping track #{track.index}...", color=Colors.WARN)
				continue

			TRACK_DURATION_DELTA = 5
			SEARCH_TRACKS_COUNT = 10

			ytdlp_options = DEFAULT_YTDLP_OPTIONS | {
				"outtmpl": filepath + ".%(ext)s",
				"match_filter": yt_dlp.utils.match_filter_func(
					f"duration>={track.duration-TRACK_DURATION_DELTA} & duration<={track.duration+TRACK_DURATION_DELTA}"
				),
				"playlistend": SEARCH_TRACKS_COUNT
			}

			def search_entries(url: str) -> str:
				return yt_dlp.YoutubeDL(ytdlp_options).extract_info(url, download=False)["entries"]

			tag_print(f"Searching for track \"{track.format(args.format)}\"...\r", end="")

			try:
				entries = search_entries(f"https://www.youtube.com/results?search_query={track.quoted_keywords}&sp=CAMSAhAB")

				if len(entries) == 0:
					entries = search_entries(f"ytsearch{SEARCH_TRACKS_COUNT}:{track.keywords}")

				if len(entries) == 0:
					raise HandledError(f"No results found for track \"{track.format(args.format)}\".")

				yt_dlp.YoutubeDL(ytdlp_options).download([entries[0]["id"]])
			except Exception as e:
				tag_print(f"Error: {e}; Skipping track #{track.index}...", color=Colors.WARN)
			else:
				tag_print(f"Successfully downloaded \"{track.format(args.format)}\"! ({index}/{len(tracklist)})")

	except KeyboardInterrupt:
		tag_print("Interrupted by user.", color=Colors.FAIL)

	except HandledError as e:
		if args.verbose:
			raise e
		else:
			tag_print(f"Error: {e}", color=Colors.FAIL)

if __name__ == "__main__":
	main()



import requests, re, string
from urllib.parse import urlparse, urlunparse, parse_qsl, urlencode, quote_plus
from spotify_dlp.utils import HandledError, tag_print, Colors

# PKCE Flow Imports
import base64, random, hashlib, http.server


PKCE_APP_CLIENT_ID = "8e70634824f842519e666d1fefa91fd0"


class Item:
	id: str
	type: str
	title: str
	authors: list[str]
	album: str
	date: str
	duration: int
	cover: str
	entry: int
	index: int

	def __init__(self, item={}):
		self.id = item.get("id")
		self.type = item.get("type")
		self.title = item.get("name")

		match self.type:
			case "track":
				self.authors = [author["name"] for author in item["artists"]]
				self.album = item["album"]["name"]
				self.date = item["album"]["release_date"]
				self.duration = item["duration_ms"] // 1000
				self.cover = item["album"]["images"][0]["url"]
				self.entry = item["track_number"]
			case "episode":
				self.authors = [item["show"]["publisher"]]
				self.album = item["show"]["name"]
				self.date = item["release_date"]
				self.duration = item["duration_ms"] // 1000
				self.cover = item["images"][0]["url"]
				self.entry = None
			case _:
				self.authors = []
				self.album = None
				self.date = None
				self.duration = None
				self.cover = None
				self.entry = None

		self.index = None


	@property
	def url(self):
		return f"https://open.spotify.com/{self.type}/{self.id}"

	@property
	def uri(self):
		return f"spotify:{self.type}:{self.id}"

	@property
	def keywords(self):
		return f"{self.title} {' '.join(self.authors)} {self.album}"

	@property
	def quoted_keywords(self):
		return quote_plus(self.keywords)


	def get_format_dict(self) -> dict:
		self_dict = self.__dict__.copy()
		self_dict["authors"] = ", ".join(self_dict["authors"])
		return self_dict

	def format(self, format: str) -> str:
		return format.format(**self.get_format_dict())

	def safe_format(self, format: str) -> str:
		return re.sub(r"[/<>:\"\\|?*]", "_", self.format(format).strip())

	def format_with_index(self, format: str) -> str:
		has_placeholder = lambda f, p: any(n == p for _, n, _, _ in string.Formatter().parse(f))
		return self.format(("" if has_placeholder(format, "index") else "{index}. ") + format)


class SpotifyAPI:
	client_id: str
	access_token: str
	refresh_token: str

	def __init__(self, client_id: str = PKCE_APP_CLIENT_ID, access_token: str = None, refresh_token: str = None):
		self.client_id = client_id
		self.access_token = access_token
		self.refresh_token = refresh_token

	@classmethod
	def from_client_credentials_flow(cls, client_id, client_secret):
		token = SpotifyAPI.token_post_request({
			"grant_type": "client_credentials",
			"client_id": client_id,
			"client_secret": client_secret,
		}).get("access_token")

		return cls(client_id, token)

	@classmethod
	def from_pkce_flow(cls, client_id: str = PKCE_APP_CLIENT_ID):
		def random_string(length: int = 16) -> str:
			CHARS = string.ascii_uppercase + string.ascii_lowercase + string.digits
			return ''.join(random.choice(CHARS) for _ in range(length))

		def generate_code_challenge(code_verifier: str) -> str:
			digest = hashlib.sha256(code_verifier.encode('utf-8')).digest()
			return base64.urlsafe_b64encode(digest).decode('utf-8').rstrip('=')

		def add_url_params(url, params):
			url_parts = list(urlparse(url))
			query = dict(parse_qsl(url_parts[4]))
			query.update(params)
			url_parts[4] = urlencode(query)
			return urlunparse(url_parts)

		class HttpSpotifyAuthHandler(http.server.BaseHTTPRequestHandler):
			auth_code: str = None

			def do_GET(self):
				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.server.auth_code = dict(parse_qsl(urlparse(self.path).query)).get("code")

				if self.server.auth_code:
					self.wfile.write(b"Authentication successful! You can close this window now.")
				else:
					self.wfile.write(b"No authentication code received. Please try again.")

			def log_message(self, format, *args):
				return

		httpd = http.server.HTTPServer(("127.0.0.1", 0), HttpSpotifyAuthHandler)
		PORT = httpd.server_address[1]

		REDIRECT_URI = f"http://127.0.0.1:{PORT}/"

		CODE_VERIFIER = random_string(64)
		AUTH_URL = add_url_params("https://accounts.spotify.com/authorize", {
			"client_id": client_id,
			"response_type": "code",
			"redirect_uri": REDIRECT_URI,
			"scope": "user-library-read",
			"code_challenge_method": "S256",
			"code_challenge": generate_code_challenge(CODE_VERIFIER),
		})

		tag_print(f"Please open the following URL in your browser to authenticate:", color=Colors.BOLD)
		print(AUTH_URL)

		httpd.handle_request()
		AUTH_CODE = httpd.auth_code

		if not AUTH_CODE:
			raise HandledError("No authentication code received. Please try again.")

		content = SpotifyAPI.token_post_request({
			"client_id": client_id,
			"grant_type": "authorization_code",
			"code": AUTH_CODE,
			"redirect_uri": REDIRECT_URI,
			"code_verifier": CODE_VERIFIER,
		})

		if not content.get("access_token"):
			raise HandledError("No access token received. Please try again.")

		return cls(client_id, content.get("access_token"), content.get("refresh_token"))

	def refresh_pkce_token(self):
		try:
			res = SpotifyAPI.token_post_request({
				"grant_type": "refresh_token",
				"refresh_token": self.refresh_token,
				"client_id": self.client_id
			})
		except Exception as e:
			if str(e) == "invalid_grant":
				raise HandledError("The refresh token is invalid or has expired. Please re-authenticate.") from e
			raise

		self.access_token = res.get("access_token")
		self.refresh_token = res.get("refresh_token")


	@staticmethod
	def raise_request_if_error(response: requests.Response):
		try:
			content = response.json()
		except Exception:
			raise HandledError(f"Request to {response.url} returned invalid/missing JSON with code {response.status_code}: {response.reason}")

		if "error" in content:
			if isinstance(content["error"], str):
				msg = content["error"]
			elif isinstance(content["error"], dict) and "message" in content["error"]:
				msg = content["error"]["message"]
			else:
				msg = "An unknown error occurred while processing the request."

			raise HandledError(msg)

		if response.status_code != 200:
			raise HandledError(f"Request to {response.url} failed with code {response.status_code}: {response.reason}")

	@staticmethod
	def token_post_request(data: dict = None) -> dict:
		response = requests.post(
			"https://accounts.spotify.com/api/token",
			headers={"Content-Type": "application/x-www-form-urlencoded"},
			data=urlencode(data)
		)

		SpotifyAPI.raise_request_if_error(response)

		return response.json()

	def api_get_request(self, endpoint: str) -> dict:
		response = requests.get(
			"https://api.spotify.com/v1" + endpoint,
			headers={"Authorization": "Bearer " + self.access_token}
		)

		self.raise_request_if_error(response)

		return response.json()


	def items_by_url(self, url: str) -> list[Item]:
		item_type, item_id = self.parse_url(url)
		info = []

		match item_type:
			case "album":
				album = self.api_get_request(f"/albums/{item_id}")
				while len(info) < album["total_tracks"]:
					result = self.api_get_request(f"/albums/{item_id}/tracks?limit=50&offset={len(info)}")
					for item in result["items"]:
						item["album"] = album
						info.append(Item(item))

			case "artist":
				result = self.api_get_request(f"/artists/{item_id}/top-tracks?market=US")
				for item in result["tracks"]:
					info.append(Item(item))

			case "playlist" | "saved":
				total = None
				while total is None or len(info) < total:
					result = self.api_get_request(
						f"/playlists/{item_id}/tracks?limit=100&offset={len(info)}" if item_type == "playlist" else
						f"/me/tracks?limit=50&offset={len(info)}" if item_type in ("liked", "saved") else ""
					)
					if total is None:
						total = result["total"]
					for item in result["items"]:
						if item["track"]["type"] == "track":
							info.append(Item(item["track"]))
						else:
							total -= 1
					print("\r", end="")
					tag_print(f"Fetching items... {len(info)}/{total}", end="")
				print()

			case "track":
				result = self.api_get_request(f"/tracks/{item_id}")
				info.append(Item(result))

			case _:
				raise HandledError(f"\"{item_type}\" type is not currently supported.")

		for index, item in enumerate(info, start=1):
			item.index = index

		return info


	def items_by_search(self, query: str, search_type="track") -> list[Item]:
		result = self.api_get_request(f"/search?q={query}&type={search_type}&limit=1")
		result = list(result.values())[0]["items"]

		if len(result) == 0:
			return []

		return self.items_by_url(f"spotify:{search_type}:{result[0]['id']}")


	@staticmethod
	def parse_url(url: str) -> tuple[str, str]:
		try:
			if url in ("saved", ):
				return url, None
			return re.search(r"(?:open\.spotify\.com/|spotify:)([a-z]+)(?:/|:)(\w+)", url).groups()
		except AttributeError as e:
			raise ValueError("Invalid URL.") from e
